/****************************************************************************
**
** Copyright (C) 2016 The Qt Company Ltd.
** Contact: https://www.qt.io/licensing/
**
** This file is part of the test suite of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:GPL-EXCEPT$
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3 as published by the Free Software
** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
** included in the packaging of this file. Please review the following
** information to ensure the GNU General Public License requirements will
** be met: https://www.gnu.org/licenses/gpl-3.0.html.
**
** $QT_END_LICENSE$
**
****************************************************************************/


#ifndef WAVHEADER_H
#define WAVHEADER_H

#include <QtCore/qobject.h>
#include <QtCore/qfile.h>
#include <qaudioformat.h>

/**
 * Helper class for parsing WAV file headers.
 *
 * See https://ccrma.stanford.edu/courses/422/projects/WaveFormat/
 */
class WavHeader
{
public:
    WavHeader(const QAudioFormat &format = QAudioFormat(),
              qint64 dataLength = 0);

    // Reads WAV header and seeks to start of data
    bool read(QIODevice &device);

    // Writes WAV header
    bool write(QIODevice &device);

    const QAudioFormat& format() const;
    qint64 dataLength() const;

    static qint64 headerLength();

    static bool writeDataLength(QIODevice &device, qint64 dataLength);

private:
    QAudioFormat m_format;
    qint64 m_dataLength;
};

#endif

